from typing import Sequence

from django.contrib.postgres.aggregates import ArrayAgg
from django.contrib.postgres.expressions import ArraySubquery
from django.contrib.postgres.fields import ArrayField
from django.db.models import Func, F, Q, DateTimeField, When, Case, Value, OuterRef, Exists, Subquery
from django.db.models.functions import Cast, Coalesce
from django_cte import With, CTEManager

from django_pg_rrule.fields import RruleField
from datetime import datetime


class RecurrenceManager(CTEManager):
    date_start_field = "date_start"
    date_end_field = "date_end"
    datetime_start_field = "datetime_start"
    datetime_end_field = "datetime_end"
    until_field = "rrule_until"
    amends_field = "amends"
    amended_by_field = "amended_by"

    def with_occurrences(
        self,
        start: datetime,
        end: datetime,
        expand_start: datetime | None = None,
        expand_end: datetime | None = None,
        start_qs: Q | None = None,
        additional_filter: Q | None = None,
        select_related: Sequence | None = None,
        prefetch_related: Sequence | None = None,
        order_by: Sequence | None = None,
        expand: bool | None = False,
    ):
        """Evaluate rrules and annotate all occurrences."""

        start_qs = self if start_qs is None else start_qs

        with_qs = start_qs.filter(
            (
                Q(**{f"{self.datetime_start_field}__lte": end})
                | Q(**{f"{self.date_start_field}__lte": end.date()})
            )
            & (
                Q(**{f"{self.until_field}__isnull": True})
                | Q(**{f"{self.until_field}__gte": expand_start if expand_start else start})
            )
        )
        if additional_filter:
            with_qs = with_qs.filter(additional_filter)

        annotated_with_qs = with_qs.annotate(
            odatetime=Func(
                Case(
                    When(
                        condition=Q(rrule__isnull=False),
                        then=Func(
                            Func(
                                Cast("rrule", output_field=RruleField()),
                                Case(
                                    When(**{ f"{self.datetime_start_field}__lt": expand_start }, then=expand_start),
                                    default=F(self.datetime_start_field)
                                ) if expand_start else F(self.datetime_start_field),
                                min(expand_end, end) if expand_end else end,
                                function="get_occurrences",
                                output_field=DateTimeField(),
                            ),
                            F("rdatetimes"),
                            function="ARRAY_CAT",
                            output_field=ArrayField(DateTimeField()),
                        ),
                    ),
                    default=Func(
                        Value("{}", output_field=ArrayField(DateTimeField())),
                        F(self.datetime_start_field),
                        function="ARRAY_APPEND",
                        output_field=ArrayField(DateTimeField()),
                    ),
                ),
                function="UNNEST",
            ),
            odate=Func(
                Case(
                    When(
                        condition=Q(rrule__isnull=False),
                        then=Func(
                            Func(
                                Cast("rrule", output_field=RruleField()),
                                Case(
                                    When(**{ f"{self.date_start_field}__lt": expand_start }, then=expand_start),
                                    default=F(self.date_start_field)
                                ) if expand_start else F(self.date_start_field),
                                min(expand_end, end) if expand_end else end,
                                function="get_occurrences",
                                output_field=DateTimeField(),
                            ),
                            Cast("rdates", output_field=ArrayField(DateTimeField())),
                            function="ARRAY_CAT",
                            output_field=ArrayField(DateTimeField()),
                        ),
                    ),
                    default=Func(
                        Value("{}", output_field=ArrayField(DateTimeField())),
                        Cast(self.date_start_field, output_field=DateTimeField()),
                        function="ARRAY_APPEND",
                        output_field=ArrayField(DateTimeField()),
                    ),
                ),
                function="UNNEST",
            ),
        )
        cte = With(
            annotated_with_qs.only("id"),
            name="qodatetimes",
        )
        qs = (  # Join WITH clause with actual data
            cte.join(self.model, id=cte.col.id)
            .with_cte(cte)
            # Annotate WITH clause
            .annotate(odatetime=cte.col.odatetime, odate=cte.col.odate)
        )

        expanded_qs_in_range = qs.exclude(
            # Exclude exdatetimes and exdates
            (Q(odatetime__isnull=False) & Q(exdatetimes__contains=[F("odatetime")]))
            | (Q(odate__isnull=False) & Q(exdates__contains=[F("odate")]))
        ).filter(
            # With rrule, filter recurrences
            Q(
                odatetime__lte=end,
                odatetime__gte=start - (F("datetime_end") - F("datetime_start")),
            )
            | Q(
                odate__lte=end.date(),
                odate__gte=start.date() - (F("date_end") - F("date_start")),
            )
        )

        if expand:
            qs = expanded_qs_in_range
            amending_events = qs.filter(
                **{self.amends_field: OuterRef("pk")}
            )

            qs = qs.annotate(
                amended_by_odatetimes=
                    Case(
                        When(
                            Q(**{f"{self.amended_by_field}__isnull": False, "odatetime__isnull": False}),
                            then=ArraySubquery(
                                amending_events.values("odatetime"),
                            ),
                        ),
                        default=Value([], output_field=ArrayField(DateTimeField())),
                    ),
                amended_by_odates=
                    Case(
                        When(
                            Q(**{f"{self.amended_by_field}__isnull": False, "odate__isnull": False}),
                            then=ArraySubquery(
                                amending_events.values("odate"),
                            ),
                        ),
                        default=Value([], output_field=ArrayField(DateTimeField())),
                    ),
            )
            
            qs = qs.exclude(
                (Q(odatetime__isnull=False) & Q(amended_by_odatetimes__len__gt=0) & Q(amended_by_odatetimes__contains=[F("odatetime")]))
                | (Q(odate__isnull=False) & Q(amended_by_odates__len__gt=0) & Q(amended_by_odates__contains=[F("odate")]))
            )
        else:
            qs = with_qs.filter(Q(id__in=expanded_qs_in_range.values_list("id", flat=True)))

        if select_related:
            qs = qs.select_related(*select_related)

        if prefetch_related:
            qs = qs.prefetch_related(*prefetch_related)

        if order_by:
            qs = qs.order_by(*order_by)

        if expand:
            # Hacky way to enforce RIGHT OUTER JOIN
            # Otherwise, Django always will rewrite the join_type to LEFT OUTER JOIN/INNER JOIN
            qs.query.alias_map["qodatetimes"].join_type = "RIGHT OUTER JOIN"

        return qs
